% Minicurso: Metodos de Elementos Finitos para as Eqs de Maxwell
% Saulo P. Oliveira (UFPR), 2012
%
% Metodo de elementos finitos para a equacao de Maxwell de 2a ordem 
% Elementos triangulares, base continua por partes, meio homogeneo

clear all
tic;

% Geometria da malha

tipo = input('Tecle 1 para a malha grossa e 2 para a malha fina: ');

fid     = fopen(['QuadMaxwell.' num2str(tipo) '.ele']);
aux     = fscanf(fid,'%i',3);
num_ele = aux(1);
num_vertices_ele = aux(2);

IEN    = zeros(num_vertices_ele,num_ele);
for el = 1:num_ele
  aux  = fscanf(fid,'%i',num_vertices_ele+1);
  IEN(:,el) = aux(2:num_vertices_ele+1)';
end

fid = fopen(['QuadMaxwell.' num2str(tipo) '.node']);
aux = fscanf(fid,'%i',4);
num_vertices = aux(1);

coord     = zeros(num_vertices,2);
fronteira = zeros(num_vertices,1);
for j = 1:num_vertices
  aux = fscanf(fid,'%f',4);
  coord(j,:)   = aux(2:3);
  fronteira(j) = aux(4);
end

fronteira(1:4) = 3;

% Frequencia e propriedades materiais (homogeneas)

omega   = 10;

mu      = 1.0;
epsilon = 1.0; 
sigma   = 0.0;
epstil  = epsilon + i*sigma/omega;

% Vetor com componentes da forca sobre cada vertice

x = coord(:,1);
y = coord(:,2);

M = 3;
fx =   (M^2*pi^2-omega^2*epstil)*sin(M*pi*y);
fy = i*(M^2*pi^2-omega^2*epstil)*sin(M*pi*x);

% Pontos e pesos de integracao

xch  = [0.5,0.0,0.5];
ych  = [0.0,0.5,0.5];
w    = [1,1,1]/6; 
nint = 3;

% Vetor de identificacao e no. de graus de liberdade (variaveis do sistema)
% Temos dois graus de liberdade por vertice

cnt = 0;
IDx = zeros(num_vertices,1);
IDy = zeros(num_vertices,1);
for j = 1:num_vertices
  if(fronteira(j)~=1)&(fronteira(j)~=3)
    cnt = cnt + 1;
    IDx(j) = cnt;
  end
  if(fronteira(j)~=2)&(fronteira(j)~=3)
    cnt = cnt + 1;
    IDy(j) = cnt;
  end
end
dof = cnt;

for j = 1:num_vertices  % cria "lixeira" para vertices sem graus de liberdade
  if(IDx(j)==0)
    IDx(j) = dof + 1;
  end
  if(IDy(j)==0)
    IDy(j) = dof + 1;
  end
end

% Algoritmo de montagem

K = sparse(dof+1,dof+1);
b = sparse(dof+1,1);

for el = 1:num_ele
  
% Inicializa matriz e vetor locais

  Ae = zeros(2*num_vertices_ele,2*num_vertices_ele);
  be = zeros(2*num_vertices_ele,2*num_vertices_ele);
   
  % coordenadas dos vertices do elemento
  
  ive = IEN(:,el);   % indices dos vertices do elemento
  xx  = coord(ive,1);  
  yy  = coord(ive,2);
  
  % Calculo das matrizes locais (vetoriais) por quadratura Gaussiana
  
  [N1,N2,CurlN,J] = BaseTri2DVet(xch,ych,xx,yy);
  
  for j = 1:2*num_vertices_ele 
    for l = 1:2*num_vertices_ele
      Ae(j,l) = sum( J.*w.*( CurlN(l,:).*CurlN(j,:) ) );
      be(j,l) = sum( J.*w.*( N1(l,:).*N1(j,:) + N2(l,:).*N2(j,:) ) );
    end
  end
  
  % Adiciona Ae em K
  
  II = [IDx(ive);IDy(ive)];
  K(II,II) = K(II,II) + (1/mu)*Ae - omega^2*epstil*be;
  
  % Montagem do lado direito do sistema (a fonte e' interpolada)
  
  ff = [fx(ive);fy(ive)];
  b(II) = b(II) + be*ff;

end

% solucao do sistema linear 

disp(['Tempo de montagem: ',num2str(toc)]);
tic;
uh = K(1:dof,1:dof)\b(1:dof);
uh(dof+1) = 0;
disp(['Tempo de solucao do sistema: ',num2str(toc)]);

% Prepara resultado para exibicao

Elx = zeros(3,num_ele);
Ely = zeros(3,num_ele);

for el = 1:num_ele
  Elx(:,el) = uh(IDx(IEN(:,el)));
  Ely(:,el) = uh(IDy(IEN(:,el)));
end

Elx(4,:) = Elx(3,:);   % esta definicao permite o uso das
Ely(4,:) = Ely(3,:);   % rotinas graficas para quadrilateros
IEN(4,:) = IEN(3,:);
