% Minicurso: Metodos de Elementos Finitos para as Eqs de Maxwell
% Saulo P. Oliveira (UFPR), 2012
%
% Metodo de elementos finitos para a equacao de Maxwell de 2a ordem 
% Elementos triangulares, base de Whitney, meio homogeneo

clear
tic;

% Geometria da malha

tipo = input('Tecle 1 para a malha grossa e 2 para a malha fina: ');

fid     = fopen(['QuadMaxwell.' num2str(tipo) '.ele']);
aux     = fscanf(fid,'%i',3);
num_ele = aux(1);
num_vertices_ele = aux(2);

IEN    = zeros(num_vertices_ele,num_ele);
for el = 1:num_ele
  aux  = fscanf(fid,'%i',num_vertices_ele+1);
  IEN(:,el) = aux(2:num_vertices_ele+1)';
end

fid = fopen(['QuadMaxwell.' num2str(tipo) '.node']);
aux = fscanf(fid,'%i',4);
num_vertices = aux(1);

coord     = zeros(num_vertices,2);
fronteira = zeros(num_vertices,1);
for j = 1:num_vertices
  aux = fscanf(fid,'%f',4);
  coord(j,:)   = aux(2:3);
  fronteira(j) = aux(4);
end

fid = fopen(['QuadMaxwell.' num2str(tipo) '.edge']);
aux = fscanf(fid,'%i',2);
num_edges = aux(1);

edges       = zeros(num_edges,2);
fronteira_e = zeros(num_edges,1);
for j = 1:num_edges
  aux = fscanf(fid,'%f',4);
  edges(j,:)     = aux(2:3);
  fronteira_e(j) = aux(4);
end

% associa indices de dois vertices a um unico numero concatenado 
% (ideia semelhante 'a usada no livro Jianming Jin, pag. 330)

pot10 = 10^(floor(log10(num_vertices))+1);
k1    = edges(:,1);
k2    = edges(:,2);
peso  = min( pot10*k1 + k2 , pot10*k2 + k1 ); 
peso_esparso       = sparse(pot10*num_vertices+num_vertices,1);
peso_esparso(peso) = 1:num_edges;

% cria matriz de conectividade das arestas

num_edges_ele = 3;
vc = [1,2,3,1]; % Vetor circulante sobre os vertices
IEN_e = zeros(num_edges_ele,num_ele);

for el  = 1:num_ele
  ive   = IEN(:,el); % indices dos vertices do elemento
  for j = 1:num_vertices_ele
    k1  = ive(j);       % vertice corrente
    k2  = ive(vc(j+1)); % vertice consecutivo
    pesoj1      = min( pot10*k1 + k2 , pot10*k2 + k1 ); 
    IEN_e(j,el)= peso_esparso(pesoj1);
  end
end

% Frequencia e propriedades materiais (homogeneas)

omega   = 100;

mu      = 1.0;
epsilon = 1.0; 
sigma   = 0.0;
epstil  = epsilon + i*sigma/omega;

% Vetor com componentes da forca sobre cada vertice

x = coord(:,1);
y = coord(:,2);

M = 1;
fx =   (M^2*pi^2-omega^2*epstil)*sin(M*pi*y);
fy = i*(M^2*pi^2-omega^2*epstil)*sin(M*pi*x);

% Pontos e pesos de integracao

xch  = [0.5,0.0,0.5];
ych  = [0.0,0.5,0.5];
w    = [1,1,1]/6; 
nint = 3;

% Vetor de identificacao e no. de graus de liberdade (variaveis do sistema)
% Temos dois graus de liberdade por vertice

cnt = 0;
ID = zeros(num_edges,1);
for j = 1:num_edges
  if(fronteira_e(j)==0)
    cnt = cnt + 1;
    ID(j) = cnt;
  end
end
dof = cnt;

for j = 1:num_edges  % cria "lixeira" para vertices sem graus de liberdade
  if(ID(j)==0)
    ID(j) = dof + 1;
  end
end

% Algoritmo de montagem

K = sparse(dof+1,dof+1);
b = sparse(dof+1,1);

for el = 1:num_ele
  
% Inicializa matriz e vetor locais

  Ae = zeros(num_edges_ele,num_edges_ele);
  be = zeros(num_edges_ele,num_edges_ele);
  fe = zeros(num_edges_ele,2*num_vertices_ele);
   
  % coordenadas dos vertices do elemento
  
  iee = IEN_e(:,el);  % indices das arestas do elemento
  ive = IEN(:,el);    % indices dos vertices do elemento
  xx  = coord(ive,1);  
  yy  = coord(ive,2);
  
  % Calculo das matrizes locais (vetoriais) por quadratura Gaussiana
  
  [N1p,N2p,N1,N2,CurlN,J] = BaseTri2DWit(xch,ych,xx,yy); % base de Whitney
    
  % Altera sinal da funcao de base de acordo com a orientacao do lado
  % Orientacao positiva: do menor para o maior indice (Jianming Jin, pag 332)
  
  for j = 1:num_edges_ele
    if(ive(j)>ive(vc(j+1)))
      N1(j,:)    = -N1(j,:);
      N2(j,:)    = -N2(j,:);
      CurlN(j,:) = -CurlN(j,:);
    end
  end

  for j = 1:num_edges_ele 
    for l = 1:num_edges_ele
      Ae(j,l) = sum( J.*w.*( CurlN(l,:).*CurlN(j,:) ) );
      be(j,l) = sum( J.*w.*( N1(l,:).*N1(j,:) + N2(l,:).*N2(j,:) ) );
    end
    for l = 1:2*num_vertices_ele
      fe(j,l) = sum( J.*w.*( N1p(l,:).*N1(j,:) + N2p(l,:).*N2(j,:) ) );
    end
  end
  
  % Adiciona Ae em K
  
  II = ID(iee);
  K(II,II) = K(II,II) + (1/mu)*Ae - omega^2*epstil*be;
  
  % Montagem do lado direito do sistema (a fonte e' interpolada)
  
  ff = [fx(ive);fy(ive)];
  b(II) = b(II) + fe*ff;

end

% solucao do sistema linear 

disp(['Tempo de montagem: ',num2str(toc)]);
tic;
uh = K(1:dof,1:dof)\b(1:dof);
uh(dof+1) = 0;
disp(['Tempo de solucao do sistema: ',num2str(toc)]);

% Prepara resultado para exibicao

Elx = zeros(3,num_ele);
Ely = zeros(3,num_ele);

for el = 1:num_ele
  ive = IEN(:,el); % indices dos vertices do elemento
  iee = IEN_e(:,el); % indices das arestas do elemento
  xx  = coord(ive,1);  
  yy  = coord(ive,2);
  [N1v,N2v] = BaseTri2DWitVert(xx,yy); % base de Whitney / vertices locais
  for j = 1:num_edges_ele  
    if(ive(j)>ive(vc(j+1)))  % orientacao da base (ver acima)
      N1v(j,:) = -N1v(j,:);
      N2v(j,:) = -N2v(j,:);
    end
    k = ID(iee(j));  
    Elx(:,el) = Elx(:,el) + uh(k)*N1v(j,:)';
    Ely(:,el) = Ely(:,el) + uh(k)*N2v(j,:)';
  end
end

Elx(4,:) = Elx(3,:);   % esta definicao permite o uso das
Ely(4,:) = Ely(3,:);   % rotinas graficas para quadrilateros
IEN(4,:) = IEN(3,:);
